@@ -1,39 +1,34 @@
 package HTML::Tiny;
 
+use warnings;
 use strict;
 use Carp;
+use Scalar::Util qw(blessed looks_like_number refaddr);
 
-use vars qw/$VERSION/;
-$VERSION = '0.11';
+use version; our $VERSION = qv( '0.4' );
 
 BEGIN {
 
     # http://www.w3schools.com/tags/default.asp
     for my $tag (
-        qw( a abbr acronym address area b base bdo big blockquote body br
+        qw( a abbr acronym address area b base bdo big blockquote body
         button caption cite code col colgroup dd del div dfn dl dt em
         fieldset form frame frameset h1 h2 h3 h4 h5 h6 head hr html i
-        iframe img input ins kbd label legend li link map meta noframes
-        noscript object ol optgroup option p param pre q samp script select
-        small span strong style sub sup table tbody td textarea tfoot th
-        thead title tr tt ul var )
+        iframe ins kbd label legend li link map meta noframes noscript
+        object ol optgroup option p param pre q samp script select small
+        span strong style sub sup table tbody td textarea tfoot th thead
+        title tr tt ul var )
       ) {
         no strict 'refs';
-        *$tag = sub { shift->auto_tag( $tag, @_ ) };
+        *$tag = sub { shift->tag( $tag, @_ ) };
+    }
+
+    for my $tag ( qw( br img input ) ) {
+        no strict 'refs';
+        *$tag = sub { shift->closed( $tag, @_ ) };
     }
 }
 
-# Tags that are closed (<br /> versus <br></br>)
-my @DEFAULT_CLOSED = qw( area base br col frame hr img input meta param );
-
-# Tags that get a trailing newline
-my @DEFAULT_NEWLINE = qw( html head body div p tr table );
-
-my %DEFAULT_AUTO = (
-    suffix => '',
-    method => 'tag'
-);
-
 my %ENT_MAP = (
     '&' => '&amp;',
     '<' => '&lt;',
@@ -49,52 +44,48 @@ my @UNPRINTABLE = qw(
   x18  x19  x1a  e    x1c  x1d  x1e  x1f
 );
 
+sub _hash_re {
+    my $hash = shift;
+    my $match = join( '|', map quotemeta, sort keys %$hash );
+    return qr/($match)/;
+}
+
+my $ENT_RE = _hash_re( \%ENT_MAP );
+
 sub new {
-    my $self = bless {}, shift;
-    $self->_set_auto( 'method', 'closed', @DEFAULT_CLOSED );
-    $self->_set_auto( 'suffix', "\n",     @DEFAULT_NEWLINE );
+    my $class = shift;
+    my $self = bless {}, $class;
+    $self->_initialize( @_ );
     return $self;
 }
 
-sub _set_auto {
-    my ( $self, $kind, $value ) = @_;
-
-    if ( defined $value ) {
-        for ( @_ ) {
-            $self->{autotag}->{$kind}->{$_} = $value;
-        }
-    }
-    else {
-        delete @{ $self->{autotag}->{$kind} }{@_};
-    }
+sub _initialize {
+    my $self = shift;
 }
 
 sub _str {
     my $obj = shift;
-    if ( my $ref = ref $obj ) {
-
-        # Flatten array refs...
-        return join '', @$obj
-          if 'ARRAY' eq $ref;
-
-        # ...stringify objects...
-        my $str = eval { $obj->as_string };
-        return $str unless $@;
-    }
-
+    # Flatten array refs...
+    return join '', @$obj
+      if 'ARRAY' eq ref $obj;
+    # ...stringify objects...
+    return $obj->as_string
+      if blessed $obj && $obj->can( 'as_string' );
     # ...default stringification
     return "$obj";
 }
 
 # URL encode a string
 sub url_encode {
-    my $str = _str( $_[1] );
+    my $self = shift;
+    my $str  = _str( shift );
     $str =~ s/([^A-Za-z0-9_])/$1 eq ' ' ? '+' : sprintf("%%%02x", ord($1))/eg;
     return $str;
 }
 
 sub url_decode {
-    my $str = $_[1];
+    my $self = shift;
+    my $str  = shift;
     $str =~ s/[+]/ /g;
     $str =~ s/%([0-9a-f]{2})/chr(hex($1))/eg;
     return $str;
@@ -111,22 +102,19 @@ sub query_encode {
 
 # (X)HTML entity encode a string
 sub entity_encode {
-    my $str = _str( $_[1] );
-    $str =~ s/([<>&'"])/$ENT_MAP{$1}/eg;
+    my $self = shift;
+    my $str  = _str( shift );
+    $str =~ s/$ENT_RE/$ENT_MAP{$1}/eg;
     return $str;
 }
 
 sub _tag_value {
-    my ( $self, $val ) = @_;
+    my $self = shift;
+    my $val  = shift;
     return '' if ref $val;
     return '="' . $self->entity_encode( $val ) . '"';
 }
 
-sub validate_tag {
-
-    # Do nothing. Subclass to throw an error for invalid tags
-}
-
 sub _tag {
     my $self   = shift;
     my $closed = shift;
@@ -138,8 +126,6 @@ sub _tag {
     # Merge attribute hashes
     my %attr = map { %$_ } @_;
 
-    $self->validate_tag( $closed, $name, \%attr );
-
     # Generate markup
     return "<$name"
       . join( '',
@@ -157,12 +143,10 @@ sub tag {
 
     for my $a ( @_ ) {
         if ( 'HASH' eq ref $a ) {
-
             # Merge into attributes
             %attr = ( %attr, %$a );
         }
         else {
-
             # Generate markup
             push @out,
               $self->_tag( 0, $name, \%attr )
@@ -171,10 +155,6 @@ sub tag {
         }
     }
 
-    # Special case: generate an empty tag pair if there's no content
-    push @out, $self->_tag( 0, $name, \%attr ) . $self->close( $name )
-      unless @out;
-
     return wantarray ? @out : join '', @out;
 }
 
@@ -183,20 +163,12 @@ sub closed { shift->_tag( 1, @_ ) }
 
 # Generate a closing (X)HTML tag
 sub close {
-    return "</$_[1]>";
-}
-
-sub auto_tag {
     my $self = shift;
     my $name = shift;
-    my ( $method, $post )
-      = map { $self->{autotag}->{$_}->{$name} || $DEFAULT_AUTO{$_} }
-      ( 'method', 'suffix' );
-    my @out = map { $_ . $post } $self->$method( $name, @_ );
-    return wantarray ? @out : join '', @out;
+    return "</$name>";
 }
 
-# Minimal JSON encoder. Provided here for completeness - it's useful when generating JS.
+# Minimal JSON encoder
 sub json_encode {
     my $self = shift;
     my $obj  = shift;
@@ -219,7 +191,7 @@ sub json_encode {
         }
     }
 
-    if ( $obj =~ /^-?\d+(?:[.]\d+)?$/ ) {
+    if ( looks_like_number $obj ) {
         return $obj;
     }
 
@@ -236,11 +208,11 @@ __END__
 
 =head1 NAME
 
-HTML::Tiny - Lightweight, dependency free HTML/XML generation
+HTML::Tiny - Tiny HTML generation utilities
 
 =head1 VERSION
 
-This document describes HTML::Tiny version 0.11
+This document describes HTML::Tiny version 0.4
 
 =head1 SYNOPSIS
 
@@ -408,9 +380,10 @@ would print
 
     <p><b>Foo</b><b>Bar</b></p>
 
-This behaviour is powerful but can take a little time to master. If you
-imagine '[' and ']' preventing the propagation of the 'tag individual
-items' behaviour it might help visualise how it works.
+This behaviour is powerful but can take a little time to master. If
+you imagine '[' and ']' preventing the propagation of the 'tag
+individual items' behaviour you might be close to being able to
+visualise how it works.
 
 Here's an HTML table (using the tag-name convenience methods - see
 below) that demonstrates it in more detail:
@@ -489,14 +462,7 @@ would print:
 
     <marker lat="57.0" lon="-2" />
 
-=item C<< auto_tag( $name, ... ) >>
-
-Calls either C<< tag >> or C<< closed >> based on built in rules
-for the tag. Used internally to implement the tag-named methods.
-
-=back
-
-=head2 Methods named after tags
+=item Methods named after tags
 
 In addition to the methods described above C<< HTML::Tiny >> provides
 all of the following HTML generation methods:
@@ -509,26 +475,9 @@ all of the following HTML generation methods:
     small span strong style sub sup table tbody td textarea tfoot th
     thead title tr tt ul var
 
-The following methods generate closed XHTML (<br />) tags by default:
-
-    area base br col frame hr img input meta param
-
-So:
-
-    print $h->br;   # prints <br />
-    print $h->input({ name => 'field1' });
-                    # prints <input name="field1" />
-    print $h->img({ src => 'pic.jpg' });
-                    # prints <img src="pic.jpg" />
-
-All other tag methods generate tags to wrap whatever content they
-are passed:
-
-    print $h->p('Hello, World');
-
-prints:
-
-    <p>Hello, World</p>
+With the exception of C<< br >>, C<< img >> and C<< input >> they are
+all called in the same way as C<< tag >> above - but with the tag
+name missing.
 
 So the following are equivalent:
 
@@ -538,6 +487,21 @@ and
 
     print $h->tag('a', { href => 'http://hexten.net' }, 'Hexten');
 
+C<< br >> and C<< input >> always generate closed XML style tags (in
+fact they call C<< closed >>).
+
+    print $h->br;   # prints <br />
+    print $h->input({ name => 'field1' });
+                    # prints <input name="field1" />
+    print $h->img({ src => 'pic.jpg' });
+                    # prints <img src="pic.jpg" />
+
+There's no way to override this default behaviour. If you need finer
+control over whether the tag is open or closed call C<tag>, C<open>,
+C<close> and C<closed> directly.
+
+=back
+
 =head2 Utility Methods
 
 =over
@@ -605,19 +569,6 @@ Because JSON is valid Javascript this method can be useful when generating ad-ho
 
 =back
 
-=head2 Subclassing
-
-An C<< HTML::Tiny >> is a blessed hash ref.
-
-=over
-
-=item C<< validate_tag( $closed, $name, $attr ) >>
-
-Subclass C<validate_tag> to throw an error or issue a warning when an
-attempt is made to generate an invalid tag.
-
-=back
-
 =head1 CONFIGURATION AND ENVIRONMENT
 
 HTML::Tiny requires no configuration files or environment variables.
@@ -626,8 +577,6 @@ HTML::Tiny requires no configuration files or environment variables.
 
 By design HTML::Tiny has no non-core dependencies.
 
-To run the tests you will require Test::More.
-
 =head1 INCOMPATIBILITIES
 
 None reported.
